﻿using UnityEngine;
using System.Collections.Generic;
using SimpleJSON;

namespace MacGruber
{
    public class StateMachine : MVRScript
    {
		private JSONStorableStringChooser myStateChooser;
		private JSONStorableString myCurrentStateInfo;
		private JSONStorableString myCurrentClockInfo;
		private JSONStorableString mySwitchState;
		private JSONStorableBool myPaused;
		private JSONStorableString myStateName;
		private JSONStorableBool myStateIsInitial;
		private JSONStorableBool myStateWaitForSync;
		private JSONStorableBool myStateInfiniteDuration;
		private JSONStorableFloat myStateMinDuration;
		private JSONStorableFloat myStateMaxDuration;
		private JSONStorableString myStateTransitions;
		private JSONStorableStringChooser myTransitionChooser;

		private List<object> myStateUI = new List<object>();

		private State myCurrentState = null;
		private State myInitialState = null;
		private float myClock = 0.0f;
		private int myClockInt = 0;
		private bool myNeedUIRefresh = false;

		private class State
		{
			public State(MVRScript script, string name)
			{
				myName = name;
				EnterTrigger = new EventTrigger(script, "OnEnter" + name);
				ExitTrigger = new EventTrigger(script, "OnExit" + name);
			}

			public static int SortByNameAscending(State a, State b)
			{
				return a.Name.CompareTo(b.Name);
			}

			public string Name {
				get { return myName; }
				set {
					myName = value;
					EnterTrigger.Name = "OnEnter" + value;
					ExitTrigger.Name = "OnExit" + value;
				}
			}

			public EventTrigger EnterTrigger;
			public EventTrigger ExitTrigger;
			public bool InfiniteDuration = false;
			public bool WaitForSync = false;
			public float MinDuration = 3.0f;
			public float MaxDuration = 5.0f;
			public List<State> Transitions = new List<State>();

			private string myName;
		}

		private List<State> myStates = new List<State>();

        public override void Init()
		{
			Utils.ourCreateUIElement += CreateUIElement;

			Utils.SetupInfoText(this,
				"<color=#606060><size=40><b>StateMachine</b></size>\nPowerful LogicBrick that represents a StateMachine. Each <i>State</i> has a random duration after " +
				"which it randomly chooses among the transitions of that <i>State</i> to switch the next <i>State</i>.</color>",
				250.0f, true
			);

			myCurrentStateInfo = Utils.SetupInfoOneLine(this, "Current State:", true);
			myCurrentClockInfo = Utils.SetupInfoOneLine(this, "Current Clock:", true);

			myStateChooser = new JSONStorableStringChooser("Selected State", new List<string>(), string.Empty, "Selected State");
			myStateChooser.setCallbackFunction += UISetState;
			CreateScrollablePopup(myStateChooser, false);

			Utils.SetupButton(this, "Add State", UIAddState, false);
			Utils.SetupButton(this, "Remove State", UIRemoveState, false);
			Utils.SetupButton(this, "Switch to selected State", UISwitchState, false);

			Utils.SetupAction(this, "TriggerSync", TriggerSyncAction);

			mySwitchState = new JSONStorableString("SwitchState", "", SwitchStateAction);
			mySwitchState.isStorable = mySwitchState.isRestorable = false;
			RegisterString(mySwitchState);

			myPaused = new JSONStorableBool("Time Paused (not saved)", false);
			CreateToggle(myPaused, false);

			CreateSyncSpacer(15);

			myStateName = new JSONStorableString("Name", "", UIRenameState);
			myStateIsInitial = new JSONStorableBool("Initial State", false);
			myStateIsInitial.setCallbackFunction += UISetInitialState;
			myStateInfiniteDuration = new JSONStorableBool("Infinite Duration", false);
			myStateInfiniteDuration.setCallbackFunction += UISetInfiniteDuration;
			myStateWaitForSync = new JSONStorableBool("Wait for TriggerSync", false);
			myStateWaitForSync.setCallbackFunction += UISetWaitForSync;
			myStateMinDuration = new JSONStorableFloat("Min Duration (sec)", 3, 0, 10, true, true);
			myStateMinDuration.setCallbackFunction += UISetMinDuration;
			myStateMinDuration.constrained = false;
			myStateMaxDuration = new JSONStorableFloat("Max Duration (sec)", 5, 0, 10, true, true);
			myStateMaxDuration.setCallbackFunction += UISetMaxDuration;
			myStateMaxDuration.constrained = false;

			myTransitionChooser = new JSONStorableStringChooser("Transition", new List<string>(), "", "Transition");

			SuperController.singleton.onAtomUIDRenameHandlers += OnAtomRename;
			SimpleTriggerHandler.LoadAssets();
		}

		private void UIRefresh()
		{
			UISetState(myStateChooser.val);
		}

		private void UISetState(string v)
		{
			Utils.RemoveUIElements(this, myStateUI);

			State state = GetSelectedState();
			if (state == null)
				return;

			JSONStorableString stateHeader = Utils.SetupInfoOneLine(this, "<size=30><b>State</b></size>", false);
			myStateUI.Add(stateHeader);

			myStateName.valNoCallback = state.Name;
			UIDynamicLabelInput nameInput = Utils.SetupTextInput(this, "Name", myStateName, false);
			myStateUI.Add(nameInput);

			myStateIsInitial.valNoCallback = (myInitialState == state);
			CreateToggle(myStateIsInitial, false);
			myStateUI.Add(myStateIsInitial);

			myStateInfiniteDuration.valNoCallback = state.InfiniteDuration;
			CreateToggle(myStateInfiniteDuration, false);
			myStateUI.Add(myStateInfiniteDuration);

			if (!state.InfiniteDuration)
			{
				myStateWaitForSync.valNoCallback = state.WaitForSync;
				CreateToggle(myStateWaitForSync, false);
				myStateUI.Add(myStateWaitForSync);

				myStateMinDuration.valNoCallback = state.MinDuration;
				myStateMaxDuration.valNoCallback = state.MaxDuration;
				Utils.AdjustSliderRange(myStateMinDuration);
				Utils.AdjustSliderRange(myStateMaxDuration);
				UIDynamicSlider sliderMin = CreateSlider(myStateMinDuration, false);
				UIDynamicSlider sliderMax = CreateSlider(myStateMaxDuration, false);
				sliderMin.rangeAdjustEnabled = true;
				sliderMax.rangeAdjustEnabled = true;
				myStateUI.Add(myStateMinDuration);
				myStateUI.Add(myStateMaxDuration);
			}

			UIDynamicButton enterButton = Utils.SetupButton(this, "Actions " + state.EnterTrigger.Name, state.EnterTrigger.OpenPanel, false);
			UIDynamicButton exitButton = Utils.SetupButton(this, "Actions " + state.ExitTrigger.Name, state.ExitTrigger.OpenPanel, false);
			myStateUI.Add(enterButton);
			myStateUI.Add(exitButton);

			if (myStates.Count > 1)
			{
				JSONStorableString transitionsHeader = Utils.SetupInfoOneLine(this, "<size=30><b>Transitions</b></size>", true);
				myStateUI.Add(transitionsHeader);

				List<string> transitions = myTransitionChooser.choices;
				transitions.Clear();
				for (int s=0; s<myStates.Count; ++s)
				{
					if (myStates[s] != state)
						transitions.Add(myStates[s].Name);
				}
				myTransitionChooser.choices = null;
				myTransitionChooser.choices = transitions;
				if (transitions.Count == 0)
					myTransitionChooser.val = "";
				else if (!transitions.Contains(myTransitionChooser.val))
					myTransitionChooser.val = transitions[0];
				UIDynamicPopup transitionPopup = CreateScrollablePopup(myTransitionChooser, true);
				myStateUI.Add(transitionPopup);

				UIDynamicButton transitionButton = Utils.SetupButton(this, "Add Transition", () => UIAddTransition(state), true);
				myStateUI.Add(transitionButton);

				for (int t=0; t<state.Transitions.Count; ++t)
				{
					State transition = state.Transitions[t];
					UIDynamicLabelXButton label = Utils.SetupLabelXButton(this, transition.Name, () => UIRemoveTransition(state, transition), true);
					myStateUI.Add(label);
				}
			}
		}

		private void UISetInfiniteDuration(bool v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.InfiniteDuration = v;
			UIRefresh();
		}

		private void UISetWaitForSync(bool v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.WaitForSync = v;
		}

		private void UISetMinDuration(float v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.MinDuration = v;
			Utils.AdjustMaxSliderFromMin(v, myStateMaxDuration);
			state.MaxDuration = myStateMaxDuration.val;
		}

		private void UISetMaxDuration(float v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.MaxDuration = v;
			state.MinDuration = Mathf.Min(v, state.MinDuration);
			myStateMinDuration.valNoCallback = state.MinDuration;
		}

		private void UIAddState()
		{
			for (int i=1; i<1000; ++i)
			{
				string name = "State" + i;
				if (!myStateChooser.choices.Contains(name))
				{
					myStates.Add(new State(this, name));
					myStates.Sort(State.SortByNameAscending);

					if (myInitialState == null && myStates.Count == 1)
						myInitialState = myStates[0];

					List<string> options = myStateChooser.choices;
					options.Clear();
					for (int e=0; e<myStates.Count; ++e)
						options.Add(myStates[e].Name);
					myStateChooser.choices = null; // force UI sync
					myStateChooser.choices = options;
					myStateChooser.val = name;
					return;
				}
			}
			SuperController.LogError("StateMachine: Too many states!");
		}

		private void UIRemoveState()
		{
			int idx = myStateChooser.choices.IndexOf(myStateChooser.val);
			if (idx < 0)
				return;

			State state = myStates[idx];
			if (myInitialState == state)
				myInitialState = null;
			if (myCurrentState == state)
				SwitchState(null);
			for (int s=0; s<myStates.Count; ++s)
				myStates[s].Transitions.Remove(state);

			state.EnterTrigger.Remove();
			state.ExitTrigger.Remove();
			myStates.RemoveAt(idx);

			List<string> options = myStateChooser.choices;
			options.RemoveAt(idx);
			myStateChooser.choices = null; // force UI sync
			myStateChooser.choices = options;

			if (idx < options.Count)
				myStateChooser.val = options[idx];
			else if (options.Count > 0)
				myStateChooser.val = options[options.Count-1];
			else
				myStateChooser.val = "";
		}

		private void UISwitchState()
		{
			SwitchState(GetSelectedState());
		}

		private void UIRenameState(string name)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			state.Name = name;
			myStates.Sort(State.SortByNameAscending);

			List<string> options = myStateChooser.choices;
			options.Clear();
			for (int e=0; e<myStates.Count; ++e)
				options.Add(myStates[e].Name);
			myStateChooser.choices = null; // force UI sync
			myStateChooser.choices = options;
			myStateChooser.val = name;
		}

		private void UISetInitialState(bool v)
		{
			State state = GetSelectedState();
			if (state == null)
				return;

			if (v)
				myInitialState = state;
			else if (myInitialState == state)
				myInitialState = null;
		}

		private void UIAddTransition(State state)
		{
			State transition = GetState(myTransitionChooser.val);
			if (transition == null)
				return;

			state.Transitions.Add(transition);
			state.Transitions.Sort(State.SortByNameAscending);
			UIRefresh();
		}

		private void UIRemoveTransition(State state, State transition)
		{
			state.Transitions.Remove(transition);
			UIRefresh();
		}

		private void CreateSyncSpacer(float height)
		{
			float SPACING = 15;
			float leftHeight = 0.0f;
			if (leftUIElements.Count >= 1)
				leftHeight += (leftUIElements.Count-1) * SPACING;
			for (int i=0; i < leftUIElements.Count; ++i)
			{
				RectTransform rt = leftUIElements[i] as RectTransform;
				if (rt != null)
					leftHeight += rt.sizeDelta.y;
			}
			float rightHeight = 0.0f;
			if (rightUIElements.Count >= 1)
				rightHeight += (rightUIElements.Count-1) * SPACING;
			for (int i=0; i < rightUIElements.Count; ++i)
			{
				RectTransform rt = rightUIElements[i] as RectTransform;
				if (rt != null)
					rightHeight += rt.sizeDelta.y;
			}

			height = Mathf.Max(height, 0.0f);
			bool extraHeight = height > 0;
			bool side = leftHeight > rightHeight;
			UIDynamic spacer = CreateSpacer(side);
			spacer.height = Mathf.Abs(leftHeight - rightHeight) + (extraHeight ? height : -SPACING);
			if (extraHeight)
			{
				spacer = CreateSpacer(!side);
				spacer.height = height;
			}
		}

		private void TriggerSyncAction()
		{
			if (myCurrentState != null && myClock <= 0.0f && !myCurrentState.InfiniteDuration && myCurrentState.WaitForSync)
				Transition();
		}

		private void SwitchStateAction(string newState)
		{
			mySwitchState.valNoCallback = string.Empty;
			SwitchStateByName(newState);
		}

		public void SwitchStateByName(string newState)
		{
			State state = GetState(newState);
			SwitchState(state);
		}

		private void SwitchState(State newState)
		{
			if (myCurrentState != null)
				myCurrentState.ExitTrigger.Trigger();

			myCurrentState = newState;


			if (myCurrentState != null)
			{
				myCurrentState.EnterTrigger.Trigger();
				if (myCurrentState.InfiniteDuration)
					myClock = 1.0f;
				else
					myClock = UnityEngine.Random.Range(myCurrentState.MinDuration, myCurrentState.MaxDuration);
				myClockInt = -1;
			}

			UpdateCurrentStateInfo();
		}

		private void Transition()
		{
			if (myCurrentState.Transitions.Count > 0)
			{
				int idx = UnityEngine.Random.Range(0, myCurrentState.Transitions.Count);
				State transition = myCurrentState.Transitions[idx];
				SwitchState(transition);
			}
			else
			{
				SwitchState(null);
			}
		}

		private void UpdateCurrentStateInfo()
		{
			if (myCurrentState != null)
			{
				if (myClockInt < 0)
					myCurrentStateInfo.val = "Current State: " + myCurrentState.Name;

				int clockInt = Mathf.CeilToInt(myClock*10.0f);
				if (myClockInt < 0 && myCurrentState.InfiniteDuration)
					myCurrentClockInfo.val = "Current Clock: Infinite";
				else if (clockInt == 0 && myCurrentState.WaitForSync)
					myCurrentClockInfo.val = "Current Clock: Waiting for TriggerSync";
				else if (myClockInt < 0 || myClockInt != clockInt)
					myCurrentClockInfo.val = string.Format("Current Clock: {0:F1} sec", clockInt/10.0f);
				myClockInt = clockInt;
			}
			else
			{
				myCurrentStateInfo.val = "Current State: ";
				myCurrentClockInfo.val = "Current Clock: ";
			}
		}

		private void Update()
		{
			if (myNeedUIRefresh)
			{
				myNeedUIRefresh = false;
				myStateChooser.setCallbackFunction(myStateChooser.val);
			}

			if (myCurrentState == null || myPaused.val || myCurrentState.InfiniteDuration)
				return;

			myClock = Mathf.Max(myClock - Time.deltaTime, 0.0f);
			if (myClock <= 0.0f && !myCurrentState.WaitForSync)
				Transition();
			else
				UpdateCurrentStateInfo();
		}

		private void OnAtomRename(string oldid, string newid)
		{
			for (int i=0; i<myStates.Count; ++i)
			{
				myStates[i].EnterTrigger.SyncAtomNames();
				myStates[i].ExitTrigger.SyncAtomNames();
			}
		}

		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;

				JSONArray sclist = new JSONArray();
				for (int i=0; i<myStates.Count; ++i)
				{
					State state = myStates[i];
					JSONClass sc = new JSONClass();
					sc["Name"] = state.Name;
					sc["InfiniteDuration"].AsBool = state.InfiniteDuration;
					sc["WaitForSync"].AsBool = state.WaitForSync;
					sc["MinDuration"].AsFloat = state.MinDuration;
					sc["MaxDuration"].AsFloat = state.MaxDuration;
					JSONArray tlist = new JSONArray();
					for (int t=0; t<state.Transitions.Count; ++t)
						tlist.Add("", state.Transitions[t].Name);
					sc["Transitions"] = tlist;
					sc[state.EnterTrigger.Name] = state.EnterTrigger.GetJSON(base.subScenePrefix);
					sc[state.ExitTrigger.Name] = state.ExitTrigger.GetJSON(base.subScenePrefix);
					sclist.Add("", sc);
				}
				jc["States"] = sclist;

				if (myInitialState != null)
					jc["InitialState"] = myInitialState.Name;
			}

			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				for (int i=0; i<myStates.Count; ++i)
				{
					myStates[i].EnterTrigger.Remove();
					myStates[i].ExitTrigger.Remove();
				}
				myStates.Clear();

				JSONArray sclist = jc["States"].AsArray;
				for (int i=0; i<sclist.Count; ++i)
				{
					JSONClass sc = sclist[i].AsObject;
					State state = new State(this, sc["Name"]);
					state.InfiniteDuration = sc["InfiniteDuration"].AsBool;
					state.WaitForSync = sc["WaitForSync"].AsBool;
					state.MinDuration = sc["MinDuration"].AsFloat;
					state.MaxDuration = sc["MaxDuration"].AsFloat;
					state.EnterTrigger.RestoreFromJSON(sc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					state.ExitTrigger.RestoreFromJSON(sc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
					myStates.Add(state);
				}

				myStates.Sort(State.SortByNameAscending);
				List<string> options = myStateChooser.choices;
				options.Clear();
				for (int i=0; i<myStates.Count; ++i)
					options.Add(myStates[i].Name);
				myStateChooser.choices = null; // force UI sync
				myStateChooser.choices = options;

				for (int i=0; i<sclist.Count; ++i)
				{
					JSONClass sc = sclist[i].AsObject;
					State state = GetState(sc["Name"]);
					JSONArray tlist = sc["Transitions"].AsArray;
					for (int t=0; t<tlist.Count; ++t)
					{
						State transition = GetState(tlist[t]);
						state.Transitions.Add(transition);
					}
				}

				if (jc.HasKey("InitialState"))
					myInitialState = GetState(jc["InitialState"]);
				else
					myInitialState = null;
				SwitchState(myInitialState);

				if (options.Count == 0)
					myStateChooser.valNoCallback = "";
				else if (myCurrentState != null)
					myStateChooser.valNoCallback = myCurrentState.Name;
				else if (!options.Contains(myStateChooser.val))
					myStateChooser.valNoCallback = options[0];
				myNeedUIRefresh = true;
			}
		}

		private State GetSelectedState()
		{
			return GetState(myStateChooser.val);
		}

		private State GetState(string name)
		{
			int idx = myStateChooser.choices.IndexOf(name);
			if (idx < 0)
				return null;
			else
				return myStates[idx];
		}

		private void OnDestroy()
		{
			SuperController.singleton.onAtomUIDRenameHandlers -= OnAtomRename;

			for (int i=0; i<myStates.Count; ++i)
			{
				myStates[i].EnterTrigger.Remove();
				myStates[i].ExitTrigger.Remove();
			}
			myStates.Clear();

			Utils.OnDestroy();
		}
    }
}
